# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the JetStream Tokenizer API."""

import abc
from typing import Any, Tuple, Union

import numpy as np
import jax

# Class to store token ids.
ResultTokens = Any


class Tokenizer(abc.ABC):
  """Tokenizer to convert strings to token ids and vice-versa."""

  @abc.abstractmethod
  def encode(
      self, s: str, **kwargs
  ) -> Tuple[Union[jax.Array, np.ndarray], int]:
    """Tokenize a string.
    Args:
        s: String to tokenize.
        **kwargs: Additional keyword arguments.
    Returns:
        tokens: Tokenized into integers.
        true_length: Actual length of the non-padded sequence
          if padding is used.
    """

  @abc.abstractmethod
  def decode(self, token_ids: list[int], **kwargs) -> str:
    """Processess input token ids to generate a string.
    Args:
      token_ids: List of token ids.
      **kwargs: Additional keyword arguments.
    Returns:
      str: String generated from the token ids.
    """

  @property
  @abc.abstractmethod
  def pad_id(self) -> int:
    """ID of the pad token."""

  @property
  @abc.abstractmethod
  def eos_id(self) -> int:
    """ID of EOS token."""

  @property
  @abc.abstractmethod
  def bos_id(self) -> int:
    """ID of BOS token."""

  @property
  def stop_tokens(self) -> set[int]:
    """ID of the stop token."""
    return {self.eos_id, self.pad_id}



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A test request."""

from typing import Sequence

from absl import app
from absl import flags
import grpc
from jetstream.core.proto import jetstream_pb2
from jetstream.core.proto import jetstream_pb2_grpc
from jetstream.engine.token_utils import load_vocab


_SERVER = flags.DEFINE_string("server", "0.0.0.0", "server address")
_PORT = flags.DEFINE_string("port", "9000", "port to ping")
_TEXT = flags.DEFINE_string("text", "My dog is cute", "The message")
_MAX_TOKENS = flags.DEFINE_integer(
    "max_tokens", 3, "Maximum number of output/decode tokens of a sequence"
)
_NUM_SAMPLES = flags.DEFINE_integer(
    "num_samples", 1, "Number of responses to generate per request"
)
_TOKENIZER = flags.DEFINE_string(
    "tokenizer",
    None,
    "Name or path of the tokenizer (matched to the model)",
    required=True,
)
_CLIENT_SIDE_TOKENIZATION = flags.DEFINE_bool(
    "client_side_tokenization",
    False,
    "Enable client side tokenization with tokenizer.",
)


def _GetResponseAsync(
    stub: jetstream_pb2_grpc.OrchestratorStub,
    request: jetstream_pb2.DecodeRequest,
) -> None:
  """Gets an async response."""

  response = stub.Decode(request)
  outputs = {i: [] for i in range(request.num_samples)}
  for resp in response:
    # print(len(resp.stream_content.samples))
    for i, sample in enumerate(resp.stream_content.samples):
      if _CLIENT_SIDE_TOKENIZATION.value:
        outputs[i].extend(sample.token_ids)
      else:
        outputs[i].extend(sample.text)
  print(f"Prompt: {_TEXT.value}")
  for i, output in outputs.items():
    if _CLIENT_SIDE_TOKENIZATION.value:
      vocab = load_vocab(_TOKENIZER.value)
      text_output = vocab.tokenizer.decode(output)
    else:
      text_output = "".join(output)
    print(f"The {i}th Response: {text_output}")


def main(argv: Sequence[str]) -> None:
  del argv
  # Note: Uses insecure_channel only for local testing. Please add grpc
  # credentials for Production.
  address = f"{_SERVER.value}:{_PORT.value}"
  with grpc.insecure_channel(address) as channel:
    grpc.channel_ready_future(channel).result()
    stub = jetstream_pb2_grpc.OrchestratorStub(channel)
    print(f"Sending request to: {address}")
    if _CLIENT_SIDE_TOKENIZATION.value:
      vocab = load_vocab(_TOKENIZER.value)
      token_ids = vocab.tokenizer.encode(_TEXT.value)
      request = jetstream_pb2.DecodeRequest(
          token_content=jetstream_pb2.DecodeRequest.TokenContent(
              token_ids=token_ids
          ),
          max_tokens=_MAX_TOKENS.value,
          num_samples=_NUM_SAMPLES.value,
      )
    else:
      request = jetstream_pb2.DecodeRequest(
          text_content=jetstream_pb2.DecodeRequest.TextContent(
              text=_TEXT.value
          ),
          max_tokens=_MAX_TOKENS.value,
          num_samples=_NUM_SAMPLES.value,
      )
    return _GetResponseAsync(stub, request)


if __name__ == "__main__":
  app.run(main)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Prefix Cache Test"""

import prefix_cache

import pytest
import unittest
import jax
import jax.numpy as jnp

from jax.experimental import mesh_utils
from jax.sharding import Mesh, NamedSharding, PartitionSpec


def create_default_value(prefix=None, true_length=1, padded_length=1, tokens=None) -> prefix_cache.Value:
  if prefix is None:
    prefix = {"decoder": jnp.array([1, 2, 3, 4, 5, 0, 0, 0, 0, 0])}
  if tokens is None:
    tokens = (1, 2, 3, 4, 5, 0, 0, 0, 0, 0)
  return prefix_cache.Value(
      prefix=prefix,
      true_length=true_length,
      padded_length=padded_length,
      tokens=tokens,
  )


def get_byte_in_use(device_idx: int = 0) -> int:
  jax.clear_caches()
  device = jax.local_devices()[device_idx]
  memory_stats = device.memory_stats()
  if memory_stats is None:
    pytest.skip("Cannot get device memory stats. Does the test run with TPU?")

  return memory_stats["bytes_in_use"]


class ValueTest(unittest.TestCase):
  """Test for Value."""

  def test_get_the_set_value(self):
    true_length = 5
    tokens = (1, 2, 3, 4, 5, 0, 0, 0, 0, 0)
    padded_length = 10
    prefix = {
        "decoder": {
            "layer_0": {
                "cached_prefill_key": jnp.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=jnp.int8),
                "cached_prefill_value": jnp.array([11, 12, 13, 14, 15, 16, 17, 18, 19, 20], dtype=jnp.float32),
            },
            "layer_1": {
                "cached_prefill_key": jnp.array([11, 22, 33, 44, 55, 66, 77, 88, 99, 1010], dtype=jnp.float16),
                "cached_prefill_value": jnp.array(
                    [1111, 1212, 1313, 1414, 1515, 1616, 1717, 1818, 1919, 2020],
                    dtype=jnp.int32,
                ),
            },
        }
    }
    # array len * byte size(int8, + float32 + float16 + int32) in the prefix
    prefix_size_byte = 10 * (1 + 4 + 2 + 4)
    value = prefix_cache.Value(
        prefix=prefix,
        true_length=true_length,
        padded_length=padded_length,
        tokens=tokens,
    )
    assert value.true_length == true_length
    assert value.padded_length == padded_length
    assert value.tokens == tokens
    assert jax.tree_util.tree_all(jax.tree.map(jnp.array_equal, value.prefix, prefix))
    assert value.prefix_size_bytes == prefix_size_byte

  def test_set_none_prefix_without_error(self):
    value = prefix_cache.Value(
        prefix=None,
        true_length=1,
        padded_length=2,
        tokens=(1,),
    )
    assert value.prefix is None
    assert value.prefix_size_bytes == 0

  def test_throw_exception_if_prefix_containing_non_array(self):
    with self.assertRaises(Exception):
      create_default_value(prefix={"a": "abc"})

  def test_adjust_true_length_shorter_equal_than_tokens(self):
    value = create_default_value(true_length=100, tokens=jnp.array([1, 2, 3]))
    assert value.true_length == 3

  def test_equal(self):
    prefix1 = {
        "decoder": jnp.array([1, 2, 3], dtype=jnp.int8),
    }
    prefix2 = {
        "decoder": jnp.array([1, 2, 3, 4, 5], dtype=jnp.int8),
    }
    value1_1 = create_default_value(prefix=prefix1)
    value1_2 = create_default_value(prefix=prefix1)
    assert value1_1 != 42
    assert value1_1 == value1_2
    assert value1_1.true_length == value1_2.true_length
    assert value1_1.padded_length == value1_2.padded_length
    assert value1_1.tokens == value1_2.tokens
    assert jax.tree_util.tree_all(jax.tree.map(jnp.array_equal, value1_1.prefix, value1_2.prefix))
    assert value1_1.prefix_size_bytes == value1_2.prefix_size_bytes
    value1_2 = create_default_value(prefix=prefix2)
    assert value1_1 != value1_2

  @pytest.mark.tpu_only
  def test_device_saved_as_prefix_tree(self):
    local_devices = jax.local_devices()
    num_devices = jax.local_device_count()
    mesh_shape1 = (num_devices,)
    device_mesh1 = mesh_utils.create_device_mesh(mesh_shape1, devices=local_devices)
    mesh1 = Mesh(device_mesh1, axis_names=("x",))
    partition_spec1_1 = PartitionSpec("x", None)
    partition_spec1_2 = PartitionSpec(None, "x")
    sharding1_1 = NamedSharding(mesh1, partition_spec1_1)
    sharding1_2 = NamedSharding(mesh1, partition_spec1_2)

    prefix = {
        "a": jnp.ones((512, 512), device=local_devices[0]),
        "b": jnp.ones((mesh_shape1[0], 512, 512), device=sharding1_1),
        "c": jnp.ones((512, mesh_shape1[0], 512), device=sharding1_2),
    }
    expected_device = {
        "a": local_devices[0],
        "b": sharding1_1,
        "c": sharding1_2,
    }
    value = create_default_value(prefix=prefix)
    assert value.device == expected_device

  @pytest.mark.tpu_only
  def test_device_saved_as_prefix_tree_with_sharding_multiple_dimension(self):
    local_devices = jax.local_devices()
    num_devices = jax.local_device_count()
    # assume number of devices will be multiple of 2
    if num_devices % 2 != 0:
      pytest.skip("Need multiple of 2 devices on testing environment.")
    mesh_shape2 = (num_devices // 2, 2)
    device_mesh2 = mesh_utils.create_device_mesh(mesh_shape2, devices=local_devices)
    mesh2 = Mesh(device_mesh2, axis_names=("x", "y"))
    partition_spec2 = PartitionSpec("x", "y", None)
    sharding2 = NamedSharding(mesh2, partition_spec2)

    prefix = {
        "a": jnp.ones((512, 512), device=local_devices[0]),
        "d": jnp.ones((mesh_shape2[0], mesh_shape2[1], 512), device=sharding2),
    }
    expected_device = {
        "a": local_devices[0],
        "d": sharding2,
    }
    value = create_default_value(prefix=prefix)
    assert value.device == expected_device


class PrefixCacheTrieTest(unittest.TestCase):
  """Test for PrefixCacheTrie."""

  def test_get_longest_common_prefix_key(self):
    trie = prefix_cache.PrefixCacheTrie()
    key = (1, 2, 3, 4)
    trie.insert(key)
    assert trie.get_longest_common_prefix_key((1, 2, 3, 4)) == key
    assert trie.get_longest_common_prefix_key((1, 2, 3)) == key
    assert trie.get_longest_common_prefix_key((1, 2, 3, 4, 5)) == key
    assert trie.get_longest_common_prefix_key((1, 2, 6, 7)) == key
    assert trie.get_longest_common_prefix_key((2, 3, 4, 5)) is None

  def test_insert_longer_key_replace_shorter_key(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.insert((1, 2, 3, 4))
    assert trie.get_longest_common_prefix_key((1, 2, 3)) == (1, 2, 3, 4)

  def test_insert_key_with_different_suffix_will_store_another_one(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3, 4))
    trie.insert((1, 2, 3, 5))
    assert trie.get_longest_common_prefix_key((1, 2, 3, 4)) == (1, 2, 3, 4)
    assert trie.get_longest_common_prefix_key((1, 2, 3, 5)) == (1, 2, 3, 5)

  def test_insert_shorter_key_will_not_replace_longer_key(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3, 4))
    trie.insert((1, 2, 3))
    assert trie.get_longest_common_prefix_key((1, 2, 3, 4)) == (1, 2, 3, 4)

  def test_insert_multiple_key_and_get_longest_common_prefix(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.insert((1, 2, 4))
    trie.insert((11, 2, 3))
    trie.insert((11, 2))
    trie.insert((11, 3, 4))
    result1 = trie.get_longest_common_prefix_key((1, 2, 3, 4, 5))
    assert result1 is not None
    assert result1[:3] == (1, 2, 3)
    result2 = trie.get_longest_common_prefix_key((1, 2, 4, 5, 6))
    assert result2 is not None
    assert result2[:3] == (1, 2, 4)
    result3 = trie.get_longest_common_prefix_key((11, 2, 3, 4))
    assert result3 is not None
    assert result3[:3] == (11, 2, 3)
    result4 = trie.get_longest_common_prefix_key((11, 3, 6))
    assert result4 is not None and result4[:2] == (11, 3)

  def test_erase_matched_key(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.erase((1, 2, 3))
    assert trie.get_longest_common_prefix_key((1, 2, 3)) is None

  def test_erase_shorter_or_longer_key_will_not_effect(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.erase((1, 2, 3, 4))
    trie.erase((1, 2))
    assert trie.get_longest_common_prefix_key((1, 2, 3)) == (1, 2, 3)

  def test_erase_key_will_change_to_another_longest_common_prefix_key(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.insert((1, 2, 4))
    first_matched_key = trie.get_longest_common_prefix_key((1, 2))
    assert first_matched_key is not None
    trie.erase(first_matched_key)
    second_matched_key = trie.get_longest_common_prefix_key((1, 2))
    assert second_matched_key is not None
    assert second_matched_key[:2] == (1, 2)
    trie.erase(second_matched_key)
    assert trie.get_longest_common_prefix_key((1, 2)) is None

  def test_insert_after_erase_to_empty(self):
    trie = prefix_cache.PrefixCacheTrie()
    trie.insert((1, 2, 3))
    trie.erase((1, 2, 3))
    trie.insert((4, 5, 6))
    assert trie.get_longest_common_prefix_key((4, 5, 6)) == (4, 5, 6)


class BasicStorageTest(unittest.TestCase):

  def test_is_enough_space_remain(self):
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.has_enough_space(value.prefix_size_bytes) is True
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes - 1)
    assert storage.has_enough_space(value.prefix_size_bytes) is False

  def test_add_and_fetch_with_key_exactly_matched(self):
    key = (1, 2, 3)
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.add(key, value) is True
    assert storage.retrieve(key) == value

  def test_add_fail_if_not_enough_space(self):
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes * 2 - 1)
    assert storage.add((1,), value) is True
    # The second one will exceed 1 bytes
    assert storage.add((2,), value) is False
    assert storage.retrieve((2,)) is None

  def test_cannot_retrieve_not_exactly_matched_key(self):
    key = (1, 2, 3)
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.add(key, value) is True
    assert storage.retrieve((1, 2, 4)) is None
    assert storage.retrieve((1, 2)) is None

  def test_add_and_retrieve_multiple_keys(self):
    storage = prefix_cache.BasicStorage(max_size_bytes=1_000_000)
    value1 = create_default_value(tokens=(1,))
    storage.add((1,), value1)
    value2 = create_default_value(tokens=(2,))
    storage.add((2,), value2)
    assert storage.retrieve((1,)) == value1
    assert storage.retrieve((2,)) == value2

  def test_evict_return_evicted_value(self):
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    storage.add((1,), value)
    evict_value = storage.evict((1,))
    assert evict_value == value

  def test_evict_will_release_the_memory_usage_and_cannot_retrieve_and_evict_after_evict(self):
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    storage.add((1,), value)
    # memory is not enough
    assert storage.add((2,), create_default_value()) is False
    storage.evict((1,))
    assert storage.retrieve((1,)) is None
    assert storage.evict((1,)) is None
    # should add another after evict
    value2 = create_default_value()
    assert storage.add((2,), value2) is True
    assert storage.retrieve((2,)) == value2

  def test_evict_multiple_values(self):
    key1 = (1,)
    value1 = create_default_value(tokens=key1)
    key2 = (
        1,
        2,
    )
    value2 = create_default_value(tokens=key2)
    storage = prefix_cache.BasicStorage(max_size_bytes=value1.prefix_size_bytes * 2)
    storage.add(key1, value1)
    storage.add(key2, value2)
    evict_value = storage.evict(key2)
    assert evict_value == value2
    assert storage.retrieve(key2) is None
    assert storage.retrieve(key1) is not None
    evict_value = storage.evict(key1)
    assert storage.retrieve(key1) is None

  def test_get_max_size_bytes(self):
    storage = prefix_cache.BasicStorage(max_size_bytes=100)
    assert storage.get_max_size_bytes() == 100

  def test_contains_key(self):
    value = create_default_value()
    storage = prefix_cache.BasicStorage(max_size_bytes=value.prefix_size_bytes)
    key = (1,)
    assert storage.contains(key) is False
    storage.add(key, value)
    assert storage.contains(key) is True
    storage.evict(key)
    assert storage.contains(key) is False


class HBMStorageTest(unittest.TestCase):
  """Test roughly for HBMStorage while HBMStorage is a wrapper of BasicStorage."""

  def test_basic_usage(self):
    """Test basic usage of HBMStorage checking all functions work."""
    key = (1,)
    value = create_default_value()
    storage = prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.get_max_size_bytes() == value.prefix_size_bytes
    assert storage.contains(key) is False
    assert storage.has_enough_space(value.prefix_size_bytes) is True
    assert storage.add(key, value) is True
    assert storage.contains(key) is True
    assert storage.retrieve(key) == value
    # Only have one value size, and cannot afford the second.
    assert storage.has_enough_space(value.prefix_size_bytes) is False
    assert storage.evict(key) == value
    assert storage.contains(key) is False
    assert storage.retrieve(key) is None
    assert storage.evict(key) is None
    # After evict, it should have enough space
    assert storage.has_enough_space(value.prefix_size_bytes) is True
    assert storage.add(key, value) is True

  @pytest.mark.tpu_only
  def test_move_value_to_and_from_different_device(self):
    local_devices = jax.local_devices()
    if len(local_devices) < 2:
      pytest.skip("Need to test with multiple devices")

    key = (1,)

    device_0_bytes_before = get_byte_in_use(0)
    value = create_default_value(prefix={"a": jnp.ones((4, 4), device=local_devices[0])})
    device_0_bytes_create_value = get_byte_in_use(0)
    prefix_actually_used_byte = device_0_bytes_create_value - device_0_bytes_before
    # Storage on device 1
    storage = prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes, device=local_devices[1])
    device_1_bytes_before = get_byte_in_use(1)
    assert storage.add(key, value) is True
    # The value will not removed from device 0 until async saved complete.
    # Block the retrieved value to make sure the save complete.
    retrieved_back_value = storage.retrieve(key)
    assert retrieved_back_value is not None
    jax.block_until_ready(retrieved_back_value.prefix)
    del retrieved_back_value
    del value
    device_1_byte_after_add = get_byte_in_use(1)
    # The prefix value is saved in device 1, and delete in device 0
    assert device_1_byte_after_add - device_1_bytes_before == prefix_actually_used_byte
    assert device_0_bytes_before == get_byte_in_use(0)
    # Retrieve back to the original device
    retrieved_back_value = storage.retrieve(key)
    assert retrieved_back_value is not None
    assert device_0_bytes_create_value == get_byte_in_use(0)

  @pytest.mark.tpu_only
  def test_retrieve_to_specific_device(self):
    local_devices = jax.local_devices()
    if len(local_devices) < 2:
      pytest.skip("Need to test with multiple devices")

    key = (1,)

    device_0_bytes_before = get_byte_in_use(0)
    value = create_default_value(prefix={"a": jnp.ones((4, 4), device=local_devices[0])})
    device_0_bytes_create_value = get_byte_in_use(0)
    prefix_actually_used_byte = device_0_bytes_create_value - device_0_bytes_before
    # Storage without move device
    storage = prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.add(key, value) is True

    device_1_byte_before = get_byte_in_use(1)
    # Retrieve to device 1 from storage device 0
    retrieved_back_value = storage.retrieve(key, device=local_devices[1])
    assert retrieved_back_value is not None
    device_1_byte_after = get_byte_in_use(1)
    assert retrieved_back_value.prefix["a"].device == local_devices[1]
    assert device_1_byte_after - device_1_byte_before == prefix_actually_used_byte


class DRAMStorageTest(unittest.TestCase):
  """Test basic usage of DRAMStorage only since DRAMStorage is wrapper of BasicStorage with device_get and device_put."""

  def test_basic_usage(self):
    """Test basic usage of DRAMStorage checking all functions work."""
    key = (1,)
    value = create_default_value()
    storage = prefix_cache.DRAMStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.get_max_size_bytes() == value.prefix_size_bytes
    assert storage.has_enough_space(value.prefix_size_bytes) is True
    assert storage.contains(key) is False
    assert storage.add(key, value) is True
    assert storage.contains(key) is True
    assert storage.retrieve(key) == value
    # Only have one value size, and cannot afford the second.
    assert storage.has_enough_space(value.prefix_size_bytes) is False
    assert storage.evict(key) == value
    assert storage.contains(key) is False
    assert storage.retrieve(key) is None
    assert storage.evict(key) is None
    # After evict, it should have enough space
    assert storage.has_enough_space(value.prefix_size_bytes) is True
    assert storage.add(key, value) is True

  @pytest.mark.tpu_only
  def test_move_value_between_device_and_host(self):
    origin_hbm_byte = get_byte_in_use()
    key = (1,)
    value = create_default_value()
    storage = prefix_cache.DRAMStorage(max_size_bytes=value.prefix_size_bytes)
    value_on_device_hbm_byte = get_byte_in_use()
    assert storage.add(key, value) is True
    # add to cache will not copy another in HBM
    assert value_on_device_hbm_byte == get_byte_in_use()
    del value
    value_on_host_hbm_byte = get_byte_in_use()
    # after del the value on device, hbm memory should release
    assert value_on_host_hbm_byte == origin_hbm_byte
    device_value = storage.retrieve(key)
    # copy the value back to device
    assert value_on_device_hbm_byte == get_byte_in_use()
    del device_value

  @pytest.mark.tpu_only
  def test_value_retrieve_back_to_the_same_device_sharding(self):
    local_devices = jax.local_devices()
    num_devices = jax.local_device_count()
    mesh_shape = (num_devices,)
    device_mesh = mesh_utils.create_device_mesh(mesh_shape, devices=local_devices)
    mesh = Mesh(device_mesh, axis_names=("x",))
    partition_spec = PartitionSpec("x", None)
    sharding = NamedSharding(mesh, partition_spec)

    key = (1,)
    prefix = {
        "cache": jnp.ones((mesh_shape[0], 512, 512), device=sharding),
    }
    value = create_default_value(prefix=prefix)
    storage = prefix_cache.DRAMStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.add(key, value)
    del value
    retrieved_value = storage.retrieve(key)
    assert retrieved_value is not None
    assert retrieved_value.prefix["cache"].device == sharding

  @pytest.mark.tpu_only
  def test_retrieve_to_specific_device(self):
    local_devices = jax.local_devices()
    if len(local_devices) < 2:
      pytest.skip("Need to test with multiple devices")

    key = (1,)
    device_0_bytes_before = get_byte_in_use(0)
    value = create_default_value(prefix={"a": jnp.ones((4, 4), device=local_devices[0])})
    device_0_bytes_create_value = get_byte_in_use(0)
    prefix_actually_used_byte = device_0_bytes_create_value - device_0_bytes_before
    # Storage without move device
    storage = prefix_cache.DRAMStorage(max_size_bytes=value.prefix_size_bytes)
    assert storage.add(key, value) is True

    device_1_byte_before = get_byte_in_use(1)
    # Retrieve to device 1 from storage device 0
    retrieved_back_value = storage.retrieve(key, device=local_devices[1])
    assert retrieved_back_value is not None
    device_1_byte_after = get_byte_in_use(1)
    assert retrieved_back_value.prefix["a"].device == local_devices[1]
    assert device_1_byte_after - device_1_byte_before == prefix_actually_used_byte


class LRUStrategyTest(unittest.TestCase):

  def test_evict_none_if_no_use(self):
    strategy = prefix_cache.LRUStrategy()
    assert strategy.evict() is None

  def test_evict_FIFO(self):
    strategy = prefix_cache.LRUStrategy()
    strategy.use((1,))
    strategy.use((2,))
    strategy.use((3,))
    assert strategy.evict() == (1,)
    assert strategy.evict() == (2,)
    assert strategy.evict() == (3,)
    assert strategy.evict() is None

  def test_use_in_the_middle_update_to_the_last(self):
    strategy = prefix_cache.LRUStrategy()
    strategy.use((1,))
    strategy.use((2,))
    strategy.use((1,))
    assert strategy.evict() == (2,)
    assert strategy.evict() == (1,)
    assert strategy.evict() is None


class HierarchicalCacheTest(unittest.TestCase):

  def test_add_to_all_layers(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    key = (1,)
    assert cache.add(key, value)[0] is True
    assert layers[0].contains(key)
    assert layers[1].contains(key)

  def test_cannot_add_if_first_layer_max_size_is_not_enough(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes - 1),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    key = (1,)
    assert cache.add(key, value)[0] is False
    assert not layers[0].contains(key)
    assert not layers[1].contains(key)

  def test_add_evict_to_enough_space_by_lru(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    assert cache.add((1,), value)[0] is True
    assert cache.add((2,), value)[0] is True
    ok, evicted = cache.add((3,), value)
    assert ok is True
    assert len(evicted) == 1
    assert (1,) in evicted
    assert not layers[0].contains((1,))
    assert not layers[0].contains((2,))
    assert layers[0].contains((3,))
    assert not layers[1].contains((1,))
    assert layers[1].contains((2,))
    assert layers[1].contains((3,))

  def test_retrieve_will_from_all_layer_and_add_to_all_layer(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    assert cache.add((1,), value)[0] is True
    assert cache.add((2,), value)[0] is True
    assert not layers[0].contains((1,))
    assert layers[1].contains((1,))
    assert cache.retrieve((1,)) == value
    assert layers[0].contains((1,))
    assert not layers[0].contains((2,))

  def test_retrieve_not_exist_in_any_layers_return_none(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    assert cache.add((1,), value)[0] is True
    assert cache.add((2,), value)[0] is True
    assert cache.add((3,), value)[0] is True
    # Key (1,) is evicted since LRU
    assert cache.retrieve((1,)) is None

  def test_add_will_happen_to_all_layers_even_if_some_layers_already_contains_the_key(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    assert cache.add((1,), value)[0] is True
    assert cache.add((2,), value)[0] is True
    # layers[0] do not have (1,) now since LRU
    assert not layers[0].contains((1,))
    assert layers[1].contains((1,))
    ok, evicted = cache.add((1,), value)
    assert ok is True
    # (1,) is not evicted from the second layer
    assert not evicted
    assert layers[0].contains((1,))
    assert not layers[0].contains((2,))
    assert layers[1].contains((1,))
    assert layers[1].contains((2,))

  def test_lru_affect_all_layers_when_add_and_retrieve(self):
    value = create_default_value()
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 5),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    # There is a LRU queue length of 5. The first 2 will also at the first layer
    assert cache.add((1,), value)[0] is True
    assert cache.add((2,), value)[0] is True
    assert cache.retrieve((1,)) is not None
    assert cache.add((3,), value)[0] is True
    assert cache.add((4,), value)[0] is True
    assert cache.add((5,), value)[0] is True
    assert cache.retrieve((4,)) is not None
    ok, evicted = cache.add((6,), value)
    assert ok is True
    assert len(evicted) == 1
    assert (2,) in evicted
    assert cache.retrieve((3,)) is not None
    # Now [3, 6, 4, 5, 1] in LRU
    assert layers[0].contains((3,))
    assert layers[0].contains((6,))
    assert layers[1].contains((3,))
    assert layers[1].contains((6,))
    assert layers[1].contains((4,))
    assert layers[1].contains((5,))
    assert layers[1].contains((1,))
    # 2 is evicted
    assert not layers[1].contains((2,))

  @pytest.mark.tpu_only
  def test_retrieve_to_specific_device(self):
    local_devices = jax.local_devices()
    if len(local_devices) < 2:
      pytest.skip("Need to test with multiple devices")

    key = (1,)
    value = create_default_value(prefix=jnp.ones((512, 512), device=local_devices[0]))
    layers = (
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 2),
        prefix_cache.HBMStorage(max_size_bytes=value.prefix_size_bytes * 5),
    )
    cache = prefix_cache.HierarchicalCache(layers)
    assert cache.add(key, value)[0]
    retrieved_value = cache.retrieve(key, device=local_devices[1])
    assert retrieved_value is not None
    assert retrieved_value.prefix.device == local_devices[1]
    # The device in the Value remain the original before saved
    assert retrieved_value.device == local_devices[0]


class PrefixCacheTest(unittest.TestCase):

  def test_cache_miss_save_hit_load(self):
    max_bytes = 64 * 1024 * 1024 * 1024  # 64 GB
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    tokens = (1, 2, 3)
    no_matched_key = prefix_cache_inst.fetch_longest_common_prefix_key(tokens)
    # first seen prefix will not match any key
    assert no_matched_key is None
    # Use dummy cache which should be returned from prefill: cache, _ = prefill(tokens)
    kv_cache = {
        "decoder": {
            "layer_0": {
                "cached_prefill_key": jnp.array([1, 2, 3, 0, 0, 0, 0, 0], dtype=jnp.bfloat16),
                "cached_prefill_value": jnp.array([1, 2, 3, 0, 0, 0, 0, 0], dtype=jnp.bfloat16),
            },
            "layer_1": {
                "cached_prefill_key": jnp.array([1, 2, 3, 0, 0, 0, 0, 0], dtype=jnp.bfloat16),
                "cached_prefill_value": jnp.array([1, 2, 3, 0, 0, 0, 0, 0], dtype=jnp.bfloat16),
            },
        }
    }
    # Should copy kv_cache before saved if the kv_cache is used after saved.
    kv_cache_copy = jax.tree_util.tree_map(lambda x: x.copy(), kv_cache)
    saved_value = prefix_cache.Value(prefix=kv_cache_copy, true_length=len(tokens), padded_length=len(tokens), tokens=tokens)
    prefix_cache_inst.save(tuple(tokens), saved_value)

    tokens_with_common_prefix = tokens + (4, 5, 6)
    matched_key = prefix_cache_inst.fetch_longest_common_prefix_key(tokens_with_common_prefix)
    assert matched_key is not None
    assert matched_key == tokens
    loaded_value = prefix_cache_inst.load(matched_key)
    assert loaded_value == saved_value

  def test_clear_cache(self):
    tokens = (1, 2, 3)
    value = create_default_value(tokens=tokens)
    max_bytes = value.prefix_size_bytes
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    assert prefix_cache_inst.save(tokens, value) is True
    assert prefix_cache_inst.load(tokens) == value
    prefix_cache_inst.clear()
    assert prefix_cache_inst.load(tokens) is None

  def test_evict_cache_with_LRU_strategy(self):
    tokens1 = (1,)
    value1 = create_default_value(tokens=tokens1)
    tokens2 = (2,)
    value2 = create_default_value(tokens=tokens2)
    tokens3 = (3,)
    value3 = create_default_value(tokens=tokens3)
    # cache with 2 values size, and will evict with LRU
    max_bytes = value1.prefix_size_bytes * 2
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    assert prefix_cache_inst.save(tokens1, value1) is True
    assert prefix_cache_inst.save(tokens2, value2) is True
    assert prefix_cache_inst.load(tokens1) == value1
    assert prefix_cache_inst.save(tokens3, value3) is True
    assert prefix_cache_inst.load(tokens2) is None

  def test_fetch_longest_common_prefix_key_does_not_affect_LRU(self):
    tokens1 = (1,)
    value1 = create_default_value(tokens=tokens1)
    tokens2 = (2,)
    value2 = create_default_value(tokens=tokens2)
    tokens3 = (3,)
    value3 = create_default_value(tokens=tokens3)
    # cache with 2 values size, and will evict with LRU
    max_bytes = value1.prefix_size_bytes * 2
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    assert prefix_cache_inst.save(tokens1, value1) is True
    assert prefix_cache_inst.save(tokens2, value2) is True
    assert prefix_cache_inst.fetch_longest_common_prefix_key(tokens1) == tokens1
    assert prefix_cache_inst.save(tokens3, value3) is True
    assert prefix_cache_inst.fetch_longest_common_prefix_key(tokens1) is None
    assert prefix_cache_inst.load(tokens1) is None

  def test_evict_cache_until_feasible_to_new_cache(self):
    # value2 is double size of value1
    value1 = create_default_value(prefix=[jnp.array([1, 2, 3]) for _ in range(1)])
    value2 = create_default_value(prefix=[jnp.array([1, 2, 3]) for _ in range(2)])

    # Only feasible for two value1 or one value2
    max_bytes = value1.prefix_size_bytes * 2
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    assert prefix_cache_inst.save(key=(1,), value=value1) is True
    assert prefix_cache_inst.save(key=(2,), value=value1) is True
    assert prefix_cache_inst.load(key=(1,)) == value1
    assert prefix_cache_inst.load(key=(2,)) == value1
    assert prefix_cache_inst.save(key=(3,), value=value2) is True
    assert prefix_cache_inst.load(key=(1,)) is None
    assert prefix_cache_inst.load(key=(2,)) is None
    assert prefix_cache_inst.load(key=(3,)) == value2

  def test_will_not_evict_if_whole_cache_is_not_feasible(self):
    # value2 is triple size of value1
    value1 = create_default_value(prefix=[jnp.array([1, 2, 3]) for _ in range(1)])
    value2 = create_default_value(prefix=[jnp.array([1, 2, 3]) for _ in range(3)])

    # Only feasible for two value1 and never value2
    max_bytes = value1.prefix_size_bytes * 2
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
    assert prefix_cache_inst.save(key=(1,), value=value1) is True
    assert prefix_cache_inst.save(key=(2,), value=value1) is True
    assert prefix_cache_inst.load(key=(1,)) == value1
    assert prefix_cache_inst.load(key=(2,)) == value1
    assert prefix_cache_inst.save(key=(3,), value=value2) is False
    assert prefix_cache_inst.load(key=(1,)) == value1
    assert prefix_cache_inst.load(key=(2,)) == value1
    assert prefix_cache_inst.load(key=(3,)) is None

  def test_cannot_fetch_longest_common_prefix_key_while_fully_evicted_from_dram(self):
    value = create_default_value()
    max_bytes = value.prefix_size_bytes
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes * 2)
    assert (
        prefix_cache_inst.save(
            (1, 2),
            value,
        )
        is True
    )
    assert (
        prefix_cache_inst.save(
            (1, 3),
            value,
        )
        is True
    )
    # (1,2,) is only in DRAM now
    assert prefix_cache_inst.fetch_longest_common_prefix_key((1, 2, 3)) == (1, 2)
    assert (
        prefix_cache_inst.save(
            (2, 3),
            value,
        )
        is True
    )
    # (1,2,) is fully evicted now
    assert prefix_cache_inst.fetch_longest_common_prefix_key((1, 2, 3)) == (1, 3)

  @pytest.mark.tpu_only
  def test_hbm_memory_usage(self):
    """Test HBM memory change."""
    pre_bytes_in_use = get_byte_in_use()
    hbm_bytes = 1024
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=hbm_bytes, dram_bytes=hbm_bytes)
    # prefix_cache will not pre allocate the memory
    assert pre_bytes_in_use == get_byte_in_use()
    prefix = {"cached_prefill_key": jnp.array([1, 2, 3, 4], dtype=jnp.bfloat16)}
    prefix_create_bytes_in_use = get_byte_in_use()
    # memory would be allocated with chunked minimum size
    prefix_bytes = prefix_create_bytes_in_use - pre_bytes_in_use
    assert prefix_create_bytes_in_use == pre_bytes_in_use + prefix_bytes
    key = (1, 2, 3)
    prefix_cache_inst.save(key, create_default_value(prefix=prefix))
    # cache save will not copy
    assert prefix_create_bytes_in_use == get_byte_in_use()
    del prefix
    # prefix move into the cache
    assert prefix_create_bytes_in_use == get_byte_in_use()
    loaded_value = prefix_cache_inst.load(key)
    assert loaded_value is not None
    loaded_bytes_in_use = get_byte_in_use()
    # load cache will not copy from cache
    assert loaded_bytes_in_use == prefix_create_bytes_in_use
    del loaded_value
    del_loaded_bytes_in_use = get_byte_in_use()
    assert del_loaded_bytes_in_use == loaded_bytes_in_use
    prefix_cache_inst.clear()
    # clear the cache will clear the saved cache.
    clear_bytes_in_use = get_byte_in_use()
    assert clear_bytes_in_use == del_loaded_bytes_in_use - prefix_bytes
    assert prefix_cache_inst.load(key) is None

  @pytest.mark.tpu_only
  def test_cache_move_between_device_and_host_with_hierarchical_cache(self):
    value1 = create_default_value()
    prefix_size_bytes = value1.prefix_size_bytes
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=prefix_size_bytes, dram_bytes=prefix_size_bytes * 2)
    assert prefix_cache_inst.save((1,), value1)
    del value1
    first_save_bytes_in_use = get_byte_in_use()
    value2 = create_default_value()
    assert prefix_size_bytes == value2.prefix_size_bytes
    assert prefix_cache_inst.save((2,), value2)
    del value2
    # the first saved cache (1,) is evicted since the size of HBM is one value.
    assert first_save_bytes_in_use == get_byte_in_use()

    # loaded value will move to HBM and (2,) in the HBM layer is evicted. The loaded value is the value in the HBM layer
    loaded_value = prefix_cache_inst.load((1,))
    assert loaded_value is not None
    assert first_save_bytes_in_use == get_byte_in_use()
    del loaded_value
    assert first_save_bytes_in_use == get_byte_in_use()

    # switch the cache again
    loaded_value = prefix_cache_inst.load((2,))
    assert loaded_value is not None
    assert first_save_bytes_in_use == get_byte_in_use()
    del loaded_value
    assert first_save_bytes_in_use == get_byte_in_use()

  @pytest.mark.tpu_only
  def test_load_to_specific_device(self):
    local_devices = jax.local_devices()
    if len(local_devices) < 2:
      pytest.skip("Need to test with multiple devices")

    key = (1,)
    value = create_default_value(prefix=jnp.ones((512, 512), device=local_devices[0]))
    prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=value.prefix_size_bytes, dram_bytes=value.prefix_size_bytes * 2)
    assert prefix_cache_inst.save(key, value)
    loaded_value = prefix_cache_inst.load(key, device=local_devices[1])
    assert loaded_value is not None
    assert loaded_value.prefix.device == local_devices[1]
    assert loaded_value.device == local_devices[0]


if __name__ == "__main__":
  unittest.main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import dataclasses
import jax
from jax import numpy as jnp
from inference.nn import AttentionMetadata
from inference.utils import register_flat_dataclass_as_pytree


@register_flat_dataclass_as_pytree
@dataclasses.dataclass
class SamplingParams:
  temperature: jax.Array
  top_k: jax.Array
  rng: jax.Array


class Sampler:

  def __init__(
      self,
      eos: int | None = None,
      max_length: int | None = None,
  ) -> None:
    self.eos = eos
    self.max_length = max_length

  def sample(
      self,
      logits: jax.Array,  # [num_tokens, vocab_size]
      positions: jax.Array,
      attn_metadata: AttentionMetadata,
      sampling_params: SamplingParams,
  ) -> jax.Array:
    sampling_rng = sampling_params.rng
    probabilities = jax.nn.softmax(
        logits / sampling_params.temperature, axis=-1
    )
    top_k_prob, top_k_indices = jax.lax.top_k(probabilities, 1)
    selected_index = jax.random.categorical(sampling_rng, top_k_prob)

    tokens = top_k_indices[
        jnp.arange(0, top_k_indices.shape[0]), selected_index
    ]
    done = jnp.equal(tokens, self.eos)
    done = jnp.logical_or(
        done, jnp.greater_equal(positions, self.max_length - 1)
    )

    if len(attn_metadata.prefill_pos.shape) == 0:
      padded_prefill_len = 0
    else:
      padded_prefill_len = attn_metadata.prefill_pos.shape[0]

    if len(attn_metadata.generate_pos.shape) != 0 and padded_prefill_len != 0:
      prefill_token = tokens.at[attn_metadata.prefill_length - 1].get()[None]
      generate_tokens = tokens.at[padded_prefill_len:].get()

      prefill_done = done.at[attn_metadata.prefill_length - 1].get()[None]
      generate_done = done.at[padded_prefill_len:].get()

      tokens = jnp.concatenate((prefill_token, generate_tokens))
      done = jnp.concatenate((prefill_done, generate_done))
    elif padded_prefill_len != 0:
      tokens = tokens.at[attn_metadata.prefill_length - 1].get()[None]
      done = done.at[attn_metadata.prefill_length - 1].get()[None]

    return tokens, done



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Model registry"""

import enum
from typing import Any
from jax.sharding import Mesh
from jax import numpy as jnp
from transformers import (
    logging,
    AutoModelForCausalLM,
    AutoTokenizer,
    AutoConfig,
    PretrainedConfig,
    PreTrainedTokenizer,
)
from inference import nn
from .hf_llama_ckpt_conversion import convert_hf_llama
from .util import torch_jax_dtype_map

logging.set_verbosity_error()


@enum.unique
class ModelSource(enum.Enum):
  NATIVE = enum.auto()
  HUGGINGFACE = enum.auto()


class ModelRegistry:

  def load_tokenizer(self, model_id: str) -> PreTrainedTokenizer:
    return AutoTokenizer.from_pretrained(model_id)

  def load_model_config(self, model_id: str) -> PretrainedConfig:
    return AutoConfig.from_pretrained(model_id)

  def load_weights_to_host(
      self,
      model_id: str,
      num_devices: int,
      model_config: PretrainedConfig,
      dtype: jnp.dtype,
  ) -> Any:
    """Load the ckpt to the host DRAM."""
    hg_model = AutoModelForCausalLM.from_pretrained(
        model_id,
        torch_dtype=torch_jax_dtype_map[dtype],
        config=model_config,
    )
    state_dict = hg_model.state_dict()
    del hg_model

    weights = convert_hf_llama(state_dict, num_devices, model_config)
    return weights



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Basic Collective Operations."""

import jax
from jax import numpy as jnp
from .mesh import get_num_partitions, get_partition_index


def reduce_scatter(operand, scatter_dimension, axis_names):
  """reduce-scatter sum operation via ppermute."""
  idx = get_partition_index(axis_names=axis_names)
  num_partitions = get_num_partitions(axis_names=axis_names)
  chunk_size = operand.shape[scatter_dimension] // num_partitions
  half_chunk_size = chunk_size // 2
  half_accum_shape = (
      operand.shape[:scatter_dimension]
      + (half_chunk_size,)
      + operand.shape[scatter_dimension + 1 :]
  )

  def step(i, carry):
    accum_fwd, accum_bwd, p_fwd_res, p_bwd_res = carry
    accum_fwd += p_fwd_res
    accum_bwd += p_bwd_res

    fwd_idx = ((idx - i - 1) % num_partitions) * chunk_size
    bwd_idx = ((idx + i + 1) % num_partitions) * chunk_size + half_chunk_size
    p_fwd_res = jax.lax.dynamic_slice_in_dim(
        operand, fwd_idx, half_chunk_size, scatter_dimension
    )
    p_bwd_res = jax.lax.dynamic_slice_in_dim(
        operand, bwd_idx, half_chunk_size, scatter_dimension
    )

    accum_fwd = jax.lax.ppermute(
        accum_fwd,
        axis_name=axis_names,
        perm=[(j, (j + 1) % num_partitions) for j in range(num_partitions)],
    )
    accum_bwd = jax.lax.ppermute(
        accum_bwd,
        axis_name=axis_names,
        perm=[(j, (j - 1) % num_partitions) for j in range(num_partitions)],
    )
    return accum_fwd, accum_bwd, p_fwd_res, p_bwd_res

  accum_fwd = jnp.zeros(half_accum_shape, dtype=operand.dtype)
  accum_bwd = jnp.zeros(half_accum_shape, dtype=operand.dtype)
  initial_fwd_idx = ((idx - 1) % num_partitions) * chunk_size
  initial_bwd_idx = ((idx + 1) % num_partitions) * chunk_size + half_chunk_size
  p_fwd_res = jax.lax.dynamic_slice_in_dim(
      operand, initial_fwd_idx, half_chunk_size, scatter_dimension
  )
  p_bwd_res = jax.lax.dynamic_slice_in_dim(
      operand, initial_bwd_idx, half_chunk_size, scatter_dimension
  )

  accum_fwd, accum_bwd, p_fwd_res, p_bwd_res = jax.lax.fori_loop(
      1, num_partitions, step, (accum_fwd, accum_bwd, p_fwd_res, p_bwd_res)
  )

  return jnp.concatenate(
      (p_fwd_res, p_bwd_res), scatter_dimension
  ) + jnp.concatenate((accum_fwd, accum_bwd), scatter_dimension)


def all_reduce(operand, axis_names):
  """all-reduce sum operation"""
  return jax.lax.psum(operand, axis_name=axis_names)


def all_gather(operand, axis, axis_names):
  """all-gather operation"""
  return jax.lax.all_gather(
      operand, axis=axis, axis_name=axis_names, tiled=True
  )



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

""" Tests for kernels """

import numpy as np
from max_utils import permute_to_match_maxtext_rope, unpermute_from_match_maxtext_rope
import unittest


class HFCheckpointConversionTest(unittest.TestCase):

  def test_huggingface_to_maxtext_back_to_huggingface_flow(self):
    base_num_query_heads = 16
    head_dim = 32
    wq = np.arange(base_num_query_heads * head_dim * base_num_query_heads * head_dim, dtype=np.float16).reshape(
        base_num_query_heads * head_dim, base_num_query_heads * head_dim
    )
    wq1 = wq.transpose()
    wq2 = np.reshape(wq1, [base_num_query_heads * head_dim, base_num_query_heads, head_dim])

    wq3 = permute_to_match_maxtext_rope(wq2)
    stack_shape = (1,)
    x = np.zeros(stack_shape + wq3.shape, dtype=np.float16)
    x[0, ...] = wq3
    x = np.transpose(x, axes=(1, 0, 2, 3))

    x = x[:, 0, :, :]
    wq4 = unpermute_from_match_maxtext_rope(x, "llama3.1")
    wq5 = wq4.reshape(base_num_query_heads * head_dim, base_num_query_heads * head_dim)
    wq6 = wq5.transpose()

    if not np.array_equal(wq, wq6):
      print("Test failed: wq does not match wq6")

    if not np.array_equal(wq1, wq5):
      print("Test failed: wq1 does not match wq5")

    if not np.array_equal(wq2, wq4):
      print("Test failed: wq2 does not match wq4")


if __name__ == "__main__":
  unittest.main()



"""
Copyright 2023 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=bare-except, consider-using-generator
"""Logger that saves metrics to a local file, GCS and TensorBoard. """

import jax
import json
import os
import numpy as np

import max_logging
from utils import gcs_utils


def _prepare_metrics_for_json(metrics, step, run_name):
  """Converts metric dictionary into json supported types (e.g. float)"""
  metrics_dict = {val: float(metrics["scalar"][val]) for val in metrics["scalar"]}
  metrics_dict["step"] = float(step)
  metrics_dict["run_name"] = run_name
  return metrics_dict


class MetricLogger:
  """
  Logger for saving metrics to a local file, GCS and TensorBoard.
  """

  def __init__(self, writer, config):
    self.buffered_step = None
    self.buffered_metrics = None
    self.writer = writer
    self.config = config

  def write_metrics(self, running_gcs_metrics, metrics, step, is_training=True):
    """Entry point for all metrics writing in Train's Main.

    To avoid introducing an unnecessary dependency, we "double buffer" -- we hold
    onto the last metrics and step and only publish when we receive a new metrics and step.
    The logic is that this ensures that Jax is able to queues train_steps and we
    don't block when turning "lazy" Jax arrays into real Python numbers.
    """
    metrics_to_write, steps_to_write = None, None
    if is_training:
      if self.buffered_metrics is not None:
        if self.buffered_step is None:
          raise ValueError(f"When writing metrics, {self.buffered_step=} was none")
        metrics_to_write = self.buffered_metrics
        steps_to_write = self.buffered_step
        self.log_metrics(metrics_to_write, steps_to_write)
      self.buffered_metrics = metrics
      self.buffered_step = step
    else:
      metrics_to_write = metrics
      steps_to_write = step

    if metrics_to_write:
      if self.config.enable_tensorboard:
        self.write_metrics_to_tensorboard(metrics_to_write, steps_to_write, is_training)

      if self.config.metrics_file:
        self.write_metrics_locally(metrics_to_write, steps_to_write)

      if self.config.gcs_metrics and jax.process_index() == 0:
        running_gcs_metrics = self.write_metrics_for_gcs(metrics_to_write, steps_to_write, running_gcs_metrics, is_training)

  def log_metrics(self, metrics, step):
    """Logs metrics via max_logging"""
    max_logging.log(
        f"completed step: {step}, seconds: {metrics['scalar']['perf/step_time_seconds']:.3f}, "
        f"TFLOP/s/device: {metrics['scalar']['perf/per_device_tflops_per_sec']:.3f}, "
        f"Tokens/s/device: {metrics['scalar']['perf/per_device_tokens_per_sec']:.3f}, "
        f"total_weights: {metrics['scalar']['learning/total_weights']}, "
        f"loss: {metrics['scalar']['learning/loss']:.3f}"
    )

  def write_metrics_locally(self, metrics, step):
    """Writes metrics locally for testing"""
    with open(self.config.metrics_file, "a", encoding="utf8") as local_metrics_file:
      if step == 0:
        local_metrics_file.truncate(0)

      metrics_dict = _prepare_metrics_for_json(metrics, step, self.config.run_name)
      local_metrics_file.write(str(json.dumps(metrics_dict)) + "\n")

  def write_metrics_for_gcs(self, metrics, step, running_metrics, is_training):
    """Writes metrics to gcs"""
    metrics_dict_step = _prepare_metrics_for_json(metrics, step, self.config.run_name)
    running_metrics.append(metrics_dict_step)
    if is_training and (step + 1) % self.config.log_period == 0 or step == self.config.steps - 1:
      start_step = (step // self.config.log_period) * self.config.log_period
      metrics_filename = f"metrics_step_{start_step:06}_to_step_{step:06}.txt"
      with open(metrics_filename, "w", encoding="utf8") as metrics_for_gcs:
        for metrics_step in running_metrics:
          metrics_for_gcs.write(str(json.dumps(metrics_step)) + "\n")

      gcs_filename = os.path.join(self.config.metrics_dir, metrics_filename)
      max_logging.log(f"Moving file {metrics_filename} to GCS...")
      gcs_utils.upload_blob(gcs_filename, metrics_filename)
      max_logging.log(f"File {metrics_filename} moved successfully!")
      running_metrics = []  # reset running_metrics to empty list
    return running_metrics

  def write_metrics_to_tensorboard(self, metrics, step, is_training):
    """Writes metrics to TensorBoard"""
    with jax.spmd_mode("allow_all"):
      if jax.process_index() == 0:
        for metric_name in metrics.get("scalar", []):
          self.writer.add_scalar(metric_name, np.array(metrics["scalar"][metric_name]), step)
        for metric_name in metrics.get("scalars", []):
          self.writer.add_scalars(metric_name, metrics["scalars"][metric_name], step)

      if is_training:
        full_log = step % self.config.log_period == 0

        if full_log and jax.process_index() == 0:
          max_logging.log(f"To see full metrics 'tensorboard --logdir={self.config.tensorboard_dir}'")
          self.writer.flush()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Post-process utilities"""

import jax
from jax import numpy as jnp
import dataclasses
from jax.sharding import NamedSharding
from inference.nn import AttentionMetadata
from inference.utils import register_flat_dataclass_as_pytree


@register_flat_dataclass_as_pytree
@dataclasses.dataclass
class ModelOutput:
  prefill_token: jax.Array | NamedSharding
  prefill_done: jax.Array | NamedSharding
  prefill_next_pos: jax.Array | NamedSharding
  generate_tokens: jax.Array | NamedSharding
  generate_done: jax.Array | NamedSharding
  generate_next_pos: jax.Array | NamedSharding


def postprocess(
    tokens: jax.Array, done: jax.Array, attn_meta: AttentionMetadata
) -> ModelOutput:
  dummy_scalar = jnp.asarray(-1, dtype=jnp.int32)
  dummy_vec = jnp.asarray([-1], dtype=jnp.int32)
  output = ModelOutput(
      prefill_token=dummy_scalar,
      prefill_done=dummy_scalar,
      prefill_next_pos=dummy_scalar,
      generate_tokens=dummy_vec,
      generate_done=dummy_vec,
      generate_next_pos=dummy_vec,
  )

  has_prefill = False
  has_generate = False
  if len(attn_meta.prefill_pos.shape) != 0:
    has_prefill = True
  if len(attn_meta.generate_pos.shape) != 0:
    has_generate = True

  if has_prefill and not has_generate:
    output.prefill_token = tokens[0]
    output.prefill_done = done[0]
    output.prefill_next_pos = attn_meta.prefill_length

  if not has_prefill and has_generate:
    output.generate_tokens = tokens
    output.generate_done = done
    output.generate_next_pos = jnp.where(
        output.generate_done, -1, attn_meta.generate_pos + 1
    )
    output.generate_next_pos = jnp.where(
        output.generate_next_pos, output.generate_next_pos, -1
    )

  if has_prefill and has_generate:
    output.prefill_token = tokens[0]
    output.prefill_done = done[0]
    output.prefill_next_pos = attn_meta.prefill_length

    output.generate_tokens = tokens[1:]
    output.generate_done = done[1:]
    output.generate_next_pos = jnp.where(
        output.generate_done, -1, attn_meta.generate_pos + 1
    )
    output.generate_next_pos = jnp.where(
        output.generate_next_pos, output.generate_next_pos, -1
    )

  return output



"""
 Copyright 2025 Google LLC
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      https://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

"""Shared Benchmark config for v5e orchestrations."""

# pylint: disable=ungrouped-imports
import jax

import max_utils
import maxengine

import os
import pyconfig

from typing import Sequence
from absl import app
import jax.numpy as jnp
import datetime
from jetstream.engine import token_utils

_WARMUP_ITERS = 2
_BENCHMARK_ITERS = 5


def main(argv: Sequence[str]) -> None:
  jax.config.update("jax_default_prng_impl", "unsafe_rbg")
  os.environ["TF_CPP_MIN_LOG_LEVEL"] = "0"
  config = pyconfig.initialize(argv)
  max_utils.print_system_information()

  engine = maxengine.MaxEngine(config)
  rng = jax.random.PRNGKey(1234)
  rng, rng_load_params = jax.random.split(rng)
  params = engine.load_params(rng_load_params)

  text = config.prompt
  metadata = engine.get_tokenizer()
  tokenizer_model = engine.build_tokenizer(metadata)
  # set it to max complete prompt length that has to be bechmarked with chunked prefill
  max_prefill_length = 7000
  tokens, true_length = tokenizer_model.encode(text, is_bos=True, prefill_lengths=[max_prefill_length])

  chunk_size = 2048
  # set this to array of acceptable chunk sizes
  prefill_lengths = [1024, 2048, 4096, 7000]
  # prefill_lengths = [1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144]

  padded_tokens, true_lengths, positions = token_utils.chunk_and_pad_tokens(
      tokens, tokenizer_model.bos_id, tokenizer_model.pad_id, False, prefill_lengths, max_prefill_length, chunk_size, True
  )
  rng, _ = jax.random.split(rng)

  def run_chunked_prefill():
    prefill_result = None
    next_pos = 0
    for chunk_num, _ in enumerate(padded_tokens):
      if prefill_result is None:
        prefill_result, _ = engine.prefill(
            params=params,
            padded_tokens=padded_tokens[chunk_num],
            true_length=true_lengths[chunk_num],
            positions=positions[chunk_num],
            complete_prompt_true_length=true_length,
            complete_padded_prompt=tokens,
            previous_chunk=prefill_result,
            rng=rng,
        )
      else:
        prefill_result, _ = engine.prefill(
            params=params | {"cache": prefill_result["cache"]},
            padded_tokens=padded_tokens[chunk_num],
            true_length=true_lengths[chunk_num],
            positions=positions[chunk_num],
            complete_prompt_true_length=true_length,
            complete_padded_prompt=tokens,
            previous_chunk=prefill_result,
            rng=rng,
        )
      true_length_array = jnp.expand_dims(jnp.arange(0, chunk_num * chunk_size + true_lengths[chunk_num]), 0)
      prefill_result["true_length_array"] = true_length_array
      prefill_result["next_pos"] = jnp.full((1, 1), next_pos + true_lengths[chunk_num], dtype=jnp.int32)
      next_pos = next_pos + true_lengths[chunk_num]
    return prefill_result

  for _ in range(_WARMUP_ITERS):
    start = datetime.datetime.now()
    prefill_result = run_chunked_prefill()
    jax.block_until_ready(prefill_result)
    end = datetime.datetime.now()
    print("time taken for chunk prefill warmup: ", end - start)

  for _ in range(_BENCHMARK_ITERS):
    start = datetime.datetime.now()
    prefill_result = run_chunked_prefill()
    jax.block_until_ready(prefill_result)
    end = datetime.datetime.now()
    print("time taken for chunk prefill ", end - start)


if __name__ == "__main__":
  app.run(main)



"""
Copyright 2023 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import subprocess
import unittest

import numpy as np
import jax
from jax.sharding import Mesh
from jax.experimental import mesh_utils
from datasets import Dataset
import transformers

import pyconfig
from input_pipeline import _hf_data_processing
from input_pipeline import input_pipeline_interface

PROMPT_DATA = [
    [
        {"content": "example one question one", "role": "user"},
        {"content": "example one question two", "role": "user"},
    ],
    [
        {"content": "question two", "role": "user"},
    ],
    [
        {"content": "question three", "role": "user"},
    ],
    [
        {"content": "question four", "role": "user"},
    ],
]

COMPLETION_DATA = [
    [
        {"content": "example one answer one", "role": "assistant"},
        {"content": "example one answer two", "role": "assistant"},
    ],
    [
        {"content": "answer two", "role": "assistant"},
    ],
    [
        {"content": "answer three", "role": "assistant"},
    ],
    [
        {"content": "answer four", "role": "assistant"},
    ],
]

MESSAGES_DATA = [
    [
        {"content": "example one question one", "role": "user"},
        {"content": "example one answer one", "role": "assistant"},
        {"content": "example one question two", "role": "user"},
        {"content": "example one answer two", "role": "assistant"},
    ],
    [
        {"content": "question two", "role": "user"},
        {"content": "answer two", "role": "assistant"},
    ],
    [
        {"content": "question three", "role": "user"},
        {"content": "answer three", "role": "assistant"},
    ],
    [
        {"content": "question four", "role": "user"},
        {"content": "answer four", "role": "assistant"},
    ],
]


class SFTDataProcessingTest(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    exit_code = subprocess.call(["gsutil", "cp", "-r", "gs://maxtext-dataset/hf/llama2-tokenizer", "../assets/"])
    if exit_code != 0:
      raise ValueError(f"Download tokenizer with gsutil cp failed with exit code: {exit_code}")

  def setUp(self):
    super().setUp()
    self.config = pyconfig.initialize(
        ["sft_trainer.py", "configs/sft.yml"],
        per_device_batch_size=1,
        run_name="test",
        mesh_axes=["data"],
        logical_axis_rules=[["batch", "data"]],
        data_sharding=["data"],
        base_output_directory="gs://max-experiments/",
        tokenizer_path="../assets/llama2-tokenizer",
        train_split="train",
        enable_checkpointing=False,
        use_sft=True,
        enable_data_shuffling=False,
    )
    self.mesh_shape_1d = (len(jax.devices()),)
    self.mesh = Mesh(mesh_utils.create_device_mesh(self.mesh_shape_1d), self.config.mesh_axes)
    self.process_indices = input_pipeline_interface.get_process_loading_real_data(
        self.config.data_sharding,
        self.config.global_batch_size_to_load,
        self.config.global_batch_size_to_train_on,
        self.config.max_target_length,
        self.mesh,
    )
    self.max_target_length = 44
    self.tokenizer = transformers.AutoTokenizer.from_pretrained(
        self.config.tokenizer_path,
        add_bos_token=False,
        add_eos_token=False,
        legacy=False,
    )

  def get_train_iterator(self, train_ds, data_columns):
    self.train_iter = _hf_data_processing.preprocessing_pipeline(
        dataloading_host_index=self.process_indices.index(jax.process_index()),
        dataloading_host_count=len(self.process_indices),
        global_mesh=self.mesh,
        dataset=train_ds,
        data_column_names=data_columns,
        tokenize=self.config.tokenize_train_data,
        tokenizer_path=self.config.tokenizer_path,
        hf_access_token=self.config.hf_access_token,
        global_batch_size=self.config.global_batch_size_to_load,
        max_target_length=self.max_target_length,
        shuffle=self.config.enable_data_shuffling,
        data_shuffle_seed=self.config.data_shuffle_seed,
        add_bos=self.config.add_bos,
        add_eos=self.config.add_eos,
        packing=self.config.packing,
        generate_padding_example=False,
        use_dpo=self.config.use_dpo,
        use_sft=self.config.use_sft,
        sft_train_on_completion_only=self.config.sft_train_on_completion_only,
        grain_worker_count=0,
    )

  def test_sft_format_with_messages(self):
    train_ds = Dataset.from_dict({"messages": MESSAGES_DATA * 2})
    data_columns = ["messages"]
    self.get_train_iterator(train_ds, data_columns)

    # exp1 is longer than max_target_length=44, testing truncation
    truncated_exp1_inputs = "<s> <user>example one question one</user> <assistant>example one answer one</assistant> <user>example one question two</user> <assistant>example one answer two</assistant"
    truncated_exp1_targets = "<unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer one</assistant><unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer two</assistant<unk>"
    truncated_exp1_targets_predictable = "<unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer one</assistant><unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer two</assistant<unk>"

    # exp2 is packed from 2nd and 3rd entries, testing packing
    packed_exp2_inputs = "<s> <user>question two</user> <assistant>answer two</assistant></s><s> <user>question three</user> <assistant>answer three</assistant></s><unk><unk><unk><unk>"
    packed_exp2_targets = "<unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer two</assistant></s><s><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer three</assistant></s><unk><unk><unk><unk><unk>"
    packed_exp2_targets_predictable = "<unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer two</assistant></s><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer three</assistant></s><unk><unk><unk><unk><unk>"

    batch = next(self.train_iter)
    self.assertEqual(self.tokenizer.decode(batch["inputs"][0]), truncated_exp1_inputs)
    self.assertEqual(self.tokenizer.decode(batch["targets"][0]), truncated_exp1_targets)
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["inputs_segmentation"][0] > 0, batch["inputs"][0], 0)), truncated_exp1_inputs
    )
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["targets_segmentation"][0] > 0, batch["targets"][0], 0)),
        truncated_exp1_targets_predictable,
    )
    self.assertEqual(self.tokenizer.decode(batch["inputs"][1]), packed_exp2_inputs)
    self.assertEqual(self.tokenizer.decode(batch["targets"][1]), packed_exp2_targets)
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["inputs_segmentation"][1] > 0, batch["inputs"][1], 0)), packed_exp2_inputs
    )
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["targets_segmentation"][1] > 0, batch["targets"][1], 0)),
        packed_exp2_targets_predictable,
    )

  def test_sft_format_with_prompt_completion(self):
    train_ds = Dataset.from_dict({"prompt": PROMPT_DATA * 2, "completion": COMPLETION_DATA * 2})
    data_columns = ["prompt", "completion"]
    self.get_train_iterator(train_ds, data_columns)

    # exp1 is longer than max_target_length=44, testing truncation
    truncated_exp1_inputs = "<s> <user>example one question one</user> <assistant>example one answer one</assistant> <user>example one question two</user> <assistant>example one answer two</assistant"
    truncated_exp1_targets = "<unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer one</assistant><unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer two</assistant<unk>"
    truncated_exp1_targets_predictable = "<unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer one</assistant><unk><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>example one answer two</assistant<unk>"

    # exp2 is packed from 2nd and 3rd entries, testing packing
    packed_exp2_inputs = "<s> <user>question two</user> <assistant>answer two</assistant></s><s> <user>question three</user> <assistant>answer three</assistant></s><unk><unk><unk><unk>"
    packed_exp2_targets = "<unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer two</assistant></s><s><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer three</assistant></s><unk><unk><unk><unk><unk>"
    packed_exp2_targets_predictable = "<unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer two</assistant></s><unk><unk><unk><unk><unk><unk><unk><unk><unk> <assistant>answer three</assistant></s><unk><unk><unk><unk><unk>"

    batch = next(self.train_iter)
    self.assertEqual(self.tokenizer.decode(batch["inputs"][0]), truncated_exp1_inputs)
    self.assertEqual(self.tokenizer.decode(batch["targets"][0]), truncated_exp1_targets)
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["inputs_segmentation"][0] > 0, batch["inputs"][0], 0)), truncated_exp1_inputs
    )
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["targets_segmentation"][0] > 0, batch["targets"][0], 0)),
        truncated_exp1_targets_predictable,
    )
    self.assertEqual(self.tokenizer.decode(batch["inputs"][1]), packed_exp2_inputs)
    self.assertEqual(self.tokenizer.decode(batch["targets"][1]), packed_exp2_targets)
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["inputs_segmentation"][1] > 0, batch["inputs"][1], 0)), packed_exp2_inputs
    )
    self.assertEqual(
        self.tokenizer.decode(np.where(batch["targets_segmentation"][1] > 0, batch["targets"][1], 0)),
        packed_exp2_targets_predictable,
    )


if __name__ == "__main__":
  unittest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains common functions for a JetStream core server.

See implementations/*/sever.py for examples.
"""

import asyncio
from concurrent import futures
import gc
import logging
import os
import signal
import threading
import time
import traceback
from typing import Any, Type


import grpc
import jax
from jetstream.core import config_lib
from jetstream.core import orchestrator
from jetstream.core.metrics.prometheus import JetstreamMetricsCollector
from jetstream.core.proto import jetstream_pb2_grpc
from jetstream.engine import warmup_utils, engine_api

from prometheus_client import start_http_server

_HOST = "[::]"


class JetStreamServer:
  """JetStream grpc server."""

  def __init__(
      self, driver: orchestrator.Driver, threads: int, port, credentials
  ):
    self._executor = futures.ThreadPoolExecutor(max_workers=threads)

    self._loop = asyncio.new_event_loop()
    self._loop.set_default_executor(self._executor)
    self._loop_thread = threading.Thread(target=self._loop.run_forever)
    self._loop_thread.start()

    async def do_init():
      self._grpc_server = grpc.aio.server(
          self._executor,
      )

    asyncio.run_coroutine_threadsafe(do_init(), loop=self._loop).result()
    self._driver = driver
    jetstream_pb2_grpc.add_OrchestratorServicer_to_server(
        orchestrator.LLMOrchestrator(driver=self._driver), self._grpc_server
    )
    self._grpc_server.add_secure_port(f"{_HOST}:{port}", credentials)

  async def _async_start(self) -> None:
    await self._grpc_server.start()

  def start(self) -> None:
    asyncio.run_coroutine_threadsafe(
        self._async_start(), loop=self._loop
    ).result()

  async def _async_stop(self) -> None:
    await self._grpc_server.stop(grace=10)

  def stop(self) -> None:
    # Gracefully clean up threads in the orchestrator.
    self._driver.stop()
    asyncio.run_coroutine_threadsafe(self._async_stop(), self._loop).result()
    self._loop.call_soon_threadsafe(self._loop.stop)
    self._loop_thread.join()

  def wait_for_termination(self) -> None:
    try:
      asyncio.run_coroutine_threadsafe(
          self._grpc_server.wait_for_termination(), self._loop
      ).result()
    finally:
      self.stop()


def create_driver(
    config: Type[config_lib.ServerConfig],
    devices: Any,
    jax_padding: bool = True,
    metrics_collector: JetstreamMetricsCollector | None = None,
    enable_model_warmup: bool = False,
    multi_sampling: bool = False,
):
  """Creates a driver with a specified config.

  Args:
    config: A ServerConfig to config engine, model, device slices, etc.
    devices: Device objects, will be used to get engine with proper slicing.
    jax_padding: The flag to enable JAX padding during tokenization.
    metrics_collector: The JetStream Promethus metric collector.
    enable_model_warmup: The flag to enable model server warmup.
    multi_sampling: The flag to enable multi-sampling.

  Returns:
    An orchestrator driver.
  """
  engines = config_lib.get_engines(config, devices=devices)
  model_load_start_time = time.time()
  prefill_params = [pe.load_params() for pe in engines.prefill_engines]
  generate_params = [ge.load_params() for ge in engines.generate_engines]
  shared_params = [ie.load_params() for ie in engines.interleaved_engines]
  logging.info("Loaded all weights.")
  if metrics_collector:
    metrics_collector.get_model_load_time_metric().set(
        time.time() - model_load_start_time
    )
  interleaved_mode = (
      len(config.prefill_slices) + len(config.generate_slices) == 0
  )

  prefill_engines = engines.prefill_engines + engines.interleaved_engines
  generate_engines = engines.generate_engines + engines.interleaved_engines
  prefill_params = prefill_params + shared_params
  generate_params = generate_params + shared_params

  if prefill_engines is None:
    prefill_engines = []
  if generate_engines is None:
    generate_engines = []
  if prefill_params is None:
    prefill_params = []
  if generate_params is None:
    generate_params = []

  if enable_model_warmup:
    prefill_engines = [engine_api.JetStreamEngine(pe) for pe in prefill_engines]
    generate_engines = [
        engine_api.JetStreamEngine(ge) for ge in generate_engines
    ]

    try:
      _ = warmup_utils.layout_params_and_compile_executables(
          prefill_engines,  # pylint: disable=protected-access
          generate_engines,  # pylint: disable=protected-access
          prefill_params,  # pylint: disable=protected-access
          generate_params,  # pylint: disable=protected-access
      )

    except ValueError as e:
      print(f"Model warmup encountered an error: {e}")
      traceback.print_exc()
      os.kill(os.getpid(), signal.SIGKILL)

  return orchestrator.Driver(
      prefill_engines=prefill_engines,
      generate_engines=generate_engines,
      prefill_params=prefill_params,
      generate_params=generate_params,
      interleaved_mode=interleaved_mode,
      jax_padding=jax_padding,
      metrics_collector=metrics_collector,
      is_ray_backend=config.is_ray_backend,
      multi_sampling=multi_sampling,
  )


def run(
    port: int,
    config: Type[config_lib.ServerConfig],
    devices: Any,
    credentials: Any = grpc.insecure_server_credentials(),
    threads: int | None = None,
    jax_padding: bool = True,
    metrics_server_config: config_lib.MetricsServerConfig | None = None,
    enable_jax_profiler: bool = False,
    jax_profiler_port: int = 9999,
    enable_model_warmup: bool = False,
    multi_sampling: bool = False,
    lora_input_adapters_path: str | None = None,
) -> JetStreamServer:
  """Runs a server with a specified config.

  Args:
    port: Port on which the server will be made available.
    config: A ServerConfig to config engine, model, device slices, etc.
    devices: Device objects, will be used to get engine with proper slicing.
    credentials: Should use grpc credentials by default.
    threads: Number of RPC handlers worker threads. This should be at least
      equal to the decoding batch size to fully saturate the decoding queue.
    jax_padding: The flag to enable JAX padding during tokenization.
    metrics_server_config: The config to enable Promethus metric server.
    enable_jax_profiler: The flag to enable JAX profiler server.
    jax_profiler_port: The port JAX profiler server (default to 9999).
    enable_model_warmup: The flag to enable model server warmup.
    multi_sampling: The flag to enable multi-sampling.
    lora_input_adapters_path: Path to define the location of all lora adapters.

  Returns:
    JetStreamServer that wraps the grpc server and orchestrator driver.
  """
  # TODO: Deleting the lora_input_adapters_path for now.
  # Planning to use it in next big PR. Currently accomodating it
  # to fix the params mismatch between maxText and JetStream
  del lora_input_adapters_path

  server_start_time = time.time()
  logging.info("Kicking off gRPC server.")
  # Setup Prometheus server
  metrics_collector: JetstreamMetricsCollector = None
  if metrics_server_config and metrics_server_config.port:
    logging.info(
        "Starting Prometheus server on port %d", metrics_server_config.port
    )
    start_http_server(metrics_server_config.port)
    metrics_collector = JetstreamMetricsCollector(
        model_name=metrics_server_config.model_name
    )
  else:
    logging.info(
        "Not starting Prometheus server: --prometheus_port flag not set"
    )

  driver = create_driver(
      config,
      devices,
      jax_padding,
      metrics_collector,
      enable_model_warmup,
      multi_sampling,
  )
  # We default threads to the total number of concurrent allowed decodes,
  # to make sure we can fully saturate the model. Set default minimum to 64.
  threads = threads or max(driver.get_total_concurrent_requests(), 64)
  jetstream_server = JetStreamServer(driver, threads, port, credentials)

  # Tweak gc config.
  # Force a gen 2 collection here.
  gc.collect(generation=2)
  # Freeze objects currently tracked and ignore them in future gc runs.
  gc.freeze()
  allocs, gen1, gen2 = gc.get_threshold()
  allocs = config.gc_gen0_allocs
  gen1 = gen1 * config.gc_gen1_multipler
  gen2 = gen2 * config.gc_gen2_multipler
  gc.set_threshold(allocs, gen1, gen2)
  print("GC tweaked (allocs, gen1, gen2): ", allocs, gen1, gen2)

  logging.info("Starting server on port %d with %d threads", port, threads)
  jetstream_server.start()

  if metrics_collector:
    metrics_collector.get_server_startup_latency_metric().set(
        time.time() - server_start_time
    )

  # Setup Jax Profiler
  if enable_jax_profiler:
    logging.info("Starting JAX profiler server on port %s", jax_profiler_port)
    jax.profiler.start_server(jax_profiler_port)
  else:
    logging.info("Not starting JAX profiler server: %s", enable_jax_profiler)

  # Start profiling server by default for proxy backend.
  if jax.config.jax_platforms and "proxy" in jax.config.jax_platforms:
    from jetstream.core.utils import proxy_util  # pylint: disable=import-outside-toplevel

    thread = threading.Thread(
        target=proxy_util.start_profiling_server, args=(jax_profiler_port,)
    )
    thread.run()

  return jetstream_server


def get_devices() -> Any:
  """Gets devices."""
  # TODO: Add more logs for the devices.
  devices = jax.devices()
  logging.info("Using devices: %d", len(devices))
  return devices



# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax._src import test_util as jtu
from inference.kernel.attention.tpu import paged_attention
from inference.kernel.attention.tpu import quantization_utils
import jax.numpy as jnp
import numpy as np


jax.config.parse_flags_with_absl()


def _generate_qkv(
    seq_lens,
    page_size,
    max_seq_len,
    num_kv_heads,
    num_heads,
    head_dim,
    prng_key,
    dtype=jnp.float32,
    are_kv_quantized=False,
):
  assert max_seq_len % page_size == 0
  pages_per_sequence = max_seq_len // page_size
  batch_size = len(seq_lens)
  total_pages = batch_size * pages_per_sequence
  k1, k2, k3, k4 = jax.random.split(prng_key, 4)
  k_pages = jax.random.normal(
      k1, (num_kv_heads, total_pages, page_size, head_dim), dtype=dtype
  )
  v_pages = jax.random.normal(
      k2, (num_kv_heads, total_pages, page_size, head_dim), dtype=dtype
  )

  if are_kv_quantized:
    k_pages = quantization_utils.quantize_to_int8(k_pages)
    v_pages = quantization_utils.quantize_to_int8(v_pages)

  page_indices = jnp.arange(batch_size * pages_per_sequence, dtype=jnp.int32)
  page_indices = jax.random.permutation(k3, page_indices, independent=True)
  page_indices = page_indices.reshape(batch_size, pages_per_sequence)
  q = jax.random.normal(k4, (batch_size, num_heads, head_dim), dtype=dtype)
  return q, k_pages, v_pages, page_indices


def _reconstruct_kv(page_indices, pages):
  if isinstance(pages, quantization_utils.QuantizedTensor):
    pages = quantization_utils.unquantize_from_int8(pages, dtype=jnp.float32)

  batch_size = page_indices.shape[0]
  num_heads, _, _, head_dim = pages.shape

  def per_sequence_page_gather(pages, page_indices):
    return jnp.take(pages, page_indices, 1)

  gathered = jax.vmap(per_sequence_page_gather, in_axes=(None, 0))(
      pages, page_indices
  )
  return gathered.reshape(batch_size, num_heads, -1, head_dim)


def _grouped_query_attention_reference(q, k, v, lengths, attn_logits_soft_cap):
  batch_size, num_heads, head_dim = q.shape
  _, num_kv_heads, max_seq_len, _ = k.shape
  assert k.shape == v.shape
  assert num_heads % num_kv_heads == 0
  q = q.reshape(batch_size, num_kv_heads, num_heads // num_kv_heads, head_dim)

  if isinstance(k, quantization_utils.QuantizedTensor):
    k = quantization_utils.unquantize_from_int8(k, dtype=jnp.float32)
  if isinstance(v, quantization_utils.QuantizedTensor):
    v = quantization_utils.unquantize_from_int8(v, dtype=jnp.float32)

  logits = jnp.einsum(
      "bhgd,bhtd->bhgt", q.astype(jnp.float32), k.astype(jnp.float32)
  )
  if attn_logits_soft_cap is not None:
    logits = jnp.tanh(logits / attn_logits_soft_cap) * attn_logits_soft_cap
  mask = jnp.arange(max_seq_len)[None] < lengths[:, None]
  mask_value = -0.7 * float(np.finfo(np.dtype("float32")).max)
  logits = logits + jnp.where(mask, 0.0, mask_value)[:, None, None, :]
  weights = jax.nn.softmax(logits, axis=-1)
  o = jnp.einsum("bhgt,bhtd->bhgd", weights.astype(v.dtype), v)
  return o.reshape(batch_size, num_heads, head_dim)


def _megacore_enabled():
  return jax.devices()[0].device_kind == "TPU v4" or jtu.is_device_tpu(
      version=5, variant="p"
  )


@jtu.with_config(jax_numpy_dtype_promotion="standard")
class PagedAttentionKernelTest(jtu.JaxTestCase):

  @parameterized.product(
      dtype=(jnp.float32, jnp.bfloat16),
      page_size=(16, 32, 64),
      num_kv_heads=(1, 8),
      q_kv_head_ratio=(1, 4, 8),
      head_dim=(128, 256),
      megacore_mode=("batch", "kv_head", None),
      attn_logits_soft_cap=(1.0, None),
      are_kv_quantized=(
          False,
          True,
      ),
  )
  def test_paged_attention(
      self,
      dtype,
      page_size,
      num_kv_heads,
      q_kv_head_ratio,
      head_dim,
      megacore_mode,
      attn_logits_soft_cap,
      are_kv_quantized,
  ):
    if not jtu.is_device_tpu_at_least(4):
      self.skipTest("Only supports TPU generation 4 or above")
    if jtu.is_device_tpu(version=4) and are_kv_quantized:
      # TPU v4 has only 16MiB of VMEM which is not sufficient to store both the
      # weight and scale tensors for quantized tensors. When enabled on TPUv4,
      # the tests sometimes failed with resource exhausted error.
      self.skipTest("Quantization is not supported on TPU v4")
    if jtu.is_device_tpu_at_least(6) and are_kv_quantized:
      self.skipTest("Quantization is not supported on TPU v6")
    if megacore_mode and not _megacore_enabled():
      self.skipTest("Megacore is only available on TPU v4 or TPU v5p")
    if num_kv_heads % 2 != 0 and megacore_mode == "kv_head":
      self.skipTest("Skip kv_head megacore mode when num_kv_heads is odd")
    max_kv_len = 2048
    block_size = 512
    seq_lens = np.asarray([0, 3, 256, 513, 1023, 2048])
    q, k_pages, v_pages, page_indices = _generate_qkv(
        seq_lens,
        page_size,
        max_kv_len,
        num_kv_heads,
        num_kv_heads * q_kv_head_ratio,
        head_dim,
        jax.random.key(0),
        dtype,
        are_kv_quantized=are_kv_quantized,
    )
    o = paged_attention(
        q,
        k_pages,
        v_pages,
        seq_lens,
        page_indices,
        pages_per_compute_block=block_size // page_size,
        megacore_mode=megacore_mode,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )
    k = _reconstruct_kv(page_indices, k_pages)
    v = _reconstruct_kv(page_indices, v_pages)
    o_ref = _grouped_query_attention_reference(
        q, k, v, seq_lens, attn_logits_soft_cap
    )

    if q_kv_head_ratio > 1:
      atol, rtol = 1e-2, 2e-2
    else:
      atol, rtol = 2e-1, 1e-1
    np.testing.assert_allclose(
        o[np.where(seq_lens > 0)].astype(jnp.float32),
        o_ref[np.where(seq_lens > 0)].astype(jnp.float32),
        atol=atol,
        rtol=rtol,
    )


if __name__ == "__main__":
  absltest.main(testLoader=jtu.JaxTestLoader())



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluate accuracy of JetStream online serving."""
import re

from sympy import sympify
from sympy.core.sympify import SympifyError
from sympy.parsing.latex.errors import LaTeXParsingError
from sympy.parsing.latex import parse_latex


def replace_space_answers(text):
  return text.replace(" ", "")


def extract_numbers(string):
  # Extract numbers from the sentence, i.e. The length is 0.3 -> 0.3
  sentence = string.split(" ")
  if len(sentence) > 1:
    for word in sentence:
      if re.match(r"[+\-*/\d]+", word):
        return word
  return string


def fix_sqrt(string):
  # Fix sqrtxx -> sqrt{xx}
  string = re.sub(r"\\sqrt(-?[0-9.a-zA-Z]+)", r"\\sqrt{\1}", string)
  string = re.sub(r"\\sqrt\s+(\w+)$", r"\\sqrt{\1}", string)
  return string


def fix_a_slash_b(string):
  # Replace a/b to frac{a}{b}
  if len(string.split("/")) != 2:
    return string
  a = string.split("/")[0]
  b = string.split("/")[1]
  try:
    if "sqrt" not in a:
      a = int(a)
    if "sqrt" not in b:
      b = int(b)
    assert string == f"{a}/{b}"
    new_string = "\\frac{" + str(a) + "}{" + str(b) + "}"
    return new_string
  except ValueError:
    return string


def fix_tan(string):
  string = re.sub(r"\\tan(-?[0-9.a-zA-Z]+)", r"\\tan{\1}", string)
  string = re.sub(r"\\tan\s+(\w+)$", r"\\tan{\1}", string)
  return string


def fix_fracs(string):
  # Standardize frac to the format of frac{a}{b}
  string = string.replace("tfrac", "frac")
  string = string.replace("dfrac", "frac")
  string = string.replace("cfrac", "frac")
  substrs = string.split("\\frac")
  new_str = substrs[0]
  if len(substrs) > 1:
    substrs = substrs[1:]
    for substr in substrs:
      new_str += "\\frac"

      def _replace_match(match):
        groups = match.groups()
        return "{" + "}{".join(groups[:-1]) + "}" + groups[-1]

      patterns = [
          r"\{(\S+)\}\{(\S+)\}(.*)",  # match frac{X}{Y}
          r"(\S+)\{(\S+)\}(.*)",  # match fracX{Y}
          r"\{(\S+)\}(\S)(.*)",  # match frac{X}Y
          r"(\S)(\S)(.*)",  # match fracXY
      ]

      reformat_str = ""
      for pattern in patterns:
        match = re.match(pattern, substr)
        if match:
          reformat_str = _replace_match(match)
          break
      if reformat_str:
        new_str += reformat_str
      else:
        return string

  string = new_str
  return string


def fix_base_term(string):
  # Fix base_term, i.e. 100_8 -> 100_{8}
  match = re.match(r"(\d+)_(\d+)", string)
  if match:
    return match.group(1) + "_{" + match.group(2) + "}"
  else:
    return string


def expand_pm(string):
  # Expand pm expression to + and -
  match = re.match(r"([^,]+)\\pm([^,]+)", string)
  if match:
    plus_term = match.group(1) + "+" + match.group(2)
    mins_term = match.group(1) + "-" + match.group(2)
    return set({plus_term, mins_term})
  else:
    return set({string})


def parse_set(string):
  # Parse set expression to python list
  if isinstance(string, list):
    return string
  match = re.match(r"\\\{(.*)\\\}$", string.strip())
  if match:
    elements_string = match.group(1).strip()
    elements = [elem.strip() for elem in elements_string.split(",")]
    return elements
  else:
    return [string]


def remove_commas_from_numbers(string):
  # Remove commas within large numbers i.e., 1,234 -> 1234
  number_pattern = r"(?<!\()(\d{1,3}(?:,\d{3})+)(?!\))"
  return re.sub(
      number_pattern, lambda match: match.group(0).replace(",", ""), string
  )


def convert_leading_zero(string):
  # Add "0" if "." is the start of the string
  string = string.replace(" .", " 0.")
  string = string.replace("{.", "{0.")
  return string


def trim_general_unit(string):
  # Remove unit: miles, dollars if after is not none
  new_string = re.sub(r"\\text{.*?}$", "", string).strip()
  if new_string != "" and new_string != string:
    print(f"Warning: unit not removed: '{string}' -> '{new_string}'")
    string = new_string
  string = re.sub(r"\{(c|m)?m\}(\^(2|3))?", "", string).strip()
  string = re.sub(r"miles(\^(2|3))?", "", string).strip()
  string = re.sub(r"p\.m\.$", "", string).strip()
  string = re.sub(r"(\d)\s*t$", r"\1", string).strip()

  # Remove degree:
  string = re.sub(r"\^\\circ", "", string)

  return string


def postprocess_math(text):

  text = trim_latex_cmd(text)

  # Remove the units
  text = trim_general_unit(text)

  # Extract the answer in \text command
  match = re.search(r"\\text\{([^}]*)\}", text)
  if match:
    text = match.group(1)

  # Remove space
  text = replace_space_answers(text)

  # Standardize sqrt format as sqrt{}/{}
  text = fix_sqrt(text)

  # Standardize frac format
  text = fix_fracs(text)

  # Convert {}/{} to the format of frac {} {}
  text = fix_a_slash_b(text)

  # Handle corner cases
  text = special_handling(text)

  return text


def trim_latex_cmd(text):
  text = text.replace("\\left", "")
  text = text.replace("\\right", "")
  text = re.sub(r"\\mbox{.*?}", "", text)

  return text


def latex_matrix_to_list(latex_string):
  # Convert LaTex Matrix expression into 2d Array
  if latex_string:
    try:
      # Extract matrix content
      matrix_content = re.search(
          r"\\begin\{pmatrix\}(.*?)\\end\{pmatrix\}", latex_string, re.DOTALL
      )
      if matrix_content:
        rows = matrix_content.group(1).strip().split("\\\\")
      else:
        return None
      matrix_list = []
      for row in rows:
        cols = row.split("&")
        matrix_list.append([col.strip() for col in cols])
      return matrix_list
    except ValueError as e:
      print(f"Error parsing LaTeX matrix: {e}")
      return None

  return None


def sympify_set(text_set):
  sympified_set = set()
  for element in text_set:
    try:
      if "frac" in element:
        try:
          element = parse_latex(element)
        except LaTeXParsingError:
          pass
      sympified_set.add(sympify(element).evalf())
    except (SympifyError, AttributeError, TypeError):
      sympified_set.add(element)
  return sympified_set


def sympify_matrix(matrix):
  if matrix:
    num_rows = len(matrix)
    num_cols = len(matrix[0])

    for row in range(num_rows):
      for col in range(num_cols):
        matrix[row][col] = replace_space_answers(matrix[row][col])
        matrix[row][col] = fix_sqrt(matrix[row][col])
        matrix[row][col] = fix_fracs(matrix[row][col])
        matrix[row][col] = fix_a_slash_b(matrix[row][col])
        matrix[row][col] = parse_latex(matrix[row][col])
        matrix[row][col] = sympify(matrix[row][col])

  return matrix


def special_handling(string):

  # Remove \$ symbol for LaTex
  string = re.sub(r"\\\$", "", string)

  # Remove \th for Latex
  string = re.sub(r"\^\{\\mathrm\{th\}\}", "", string)

  # Remove , from numbers
  string = remove_commas_from_numbers(string)

  # Fix base term
  string = fix_base_term(string)
  # Expand pm or mp command
  string_set = expand_pm(string)

  # Convert LaTex format to sympy format
  try:
    string_set = parse_latex(string_set)
  except (LaTeXParsingError, AttributeError):
    pass

  return string_set


def post_processing_math_ans(answers):
  output_set = set()
  for ans in parse_set(answers):
    matrix = latex_matrix_to_list(ans)
    if matrix:
      output_set.add(str(sympify_matrix(matrix)))
    else:
      ans = postprocess_math(ans)
      output_set.update(ans)
  return output_set
